//
//  RunloopObserverSample.m
//  iOSMultiThreadSample
//
//  Created by xiaoyong.cxy on 18/07/2017.
//  Copyright © 2017 maple. All rights reserved.
//

#import "RunloopObserverSample.h"

@interface ObserverContext : NSObject

@property (nonatomic, assign) NSInteger timeoutCount;
@property (nonatomic, strong) dispatch_semaphore_t semaphore;
@property (nonatomic, assign) CFRunLoopActivity activity;

@end

@implementation ObserverContext

@end

@interface RunloopObserverSample () {
    dispatch_semaphore_t    semaphore;
    CFRunLoopObserverRef    runloopObserver;
}

@end


void runLoopObserverCallBack(CFRunLoopObserverRef observer, CFRunLoopActivity activity, void *info)
{
    ObserverContext *context = (__bridge ObserverContext *)info;
    context.activity = activity;
    dispatch_semaphore_signal(context.semaphore);
};



@implementation RunloopObserverSample

static CFRunLoopObserverRef runloopObserver = NULL;


+ (void)startRunloopObserver
{
    if (runloopObserver) {
        return;
    }
    
    ObserverContext *context = [[ObserverContext alloc] init];
    context.semaphore = dispatch_semaphore_create(0);
    
    
    CFRunLoopObserverContext observerContext = {1, (__bridge void*)context, NULL, NULL, NULL};
    
    runloopObserver = CFRunLoopObserverCreate(NULL,
                                              kCFRunLoopAllActivities,
                                              true,
                                              0,
                                              &runLoopObserverCallBack,
                                              &observerContext);
    
    CFRunLoopAddObserver(CFRunLoopGetMain(), runloopObserver, kCFRunLoopCommonModes);
    
    // 子线程监控时长
    dispatch_queue_t monitorQueue = dispatch_queue_create("MonitorQueue", DISPATCH_QUEUE_SERIAL);
    dispatch_async(monitorQueue, ^{
        while (YES)
        {
            // 假定连续5次超时50ms认为卡顿(当然也包含了单次超时250ms)
            long ret = dispatch_semaphore_wait(context.semaphore, dispatch_time(DISPATCH_TIME_NOW, 50*NSEC_PER_MSEC));
            if (ret != 0)
            {
                if (context.activity==kCFRunLoopBeforeSources
                    || context.activity==kCFRunLoopAfterWaiting)
                {
                    if (++context.timeoutCount < 5)
                        continue;
                    
                    NSLog(@"好像有点儿卡哦");
                }
            }
            context.timeoutCount = 0;
        }
    });
}

+ (void)stopRunloopObserver
{
    if (!runloopObserver)
        return;
    
    CFRunLoopRemoveObserver(CFRunLoopGetMain(), runloopObserver, kCFRunLoopCommonModes);
    CFRelease(runloopObserver);
    runloopObserver = NULL;
}

@end
