//
//  ViewController.m
//  iOSMultiThreadSample
//
//  Created by xiaoyong.cxy on 18/07/2017.
//  Copyright © 2017 maple. All rights reserved.
//

#import "ViewController.h"
#import "iOSConcurrencySample.h"
#import "SynchronizedDirectiveSample.h"
#import "RunloopObserverSample.h"

@interface CellModel : NSObject

@property (nonatomic, strong) NSString *title;
@property (nonatomic, strong) void(^actionBlock)(void);

@end


@implementation CellModel

@end


@interface ViewController () <UITableViewDelegate, UITableViewDataSource> {
    UITableView                 *_tableView;
    
    NSMutableArray <CellModel*> *_cellModels;
}

@end

@implementation ViewController

- (void)testMethod
{
    NSLog(@"haha, hello world");
}

- (void)testMethod2
{
    NSLog(@"haha, hello world 2");
}


- (void)setupCellModels
{
    _cellModels = [NSMutableArray array];
    
    CellModel *cellModel1 = [[CellModel alloc] init];
    cellModel1.title = @"Producer&Consumer with Condition";
    cellModel1.actionBlock = ^{
        [iOSConcurrencySample producerConsumerWithCondition];
    };
    
    CellModel *cellModel2 = [[CellModel alloc] init];
    cellModel2.title = @"Producer&Consumer with Condition Lock";
    cellModel2.actionBlock = ^{
        [iOSConcurrencySample producerConsumerWithConditionLock];
    };
    
    CellModel *cellModel3 = [[CellModel alloc] init];
    cellModel3.title = @"Producer&Consumer with NSLock";
    cellModel3.actionBlock = ^{
        [iOSConcurrencySample producerConsumerWithLock];
    };
    
    CellModel *cellModel4 = [[CellModel alloc] init];
    cellModel4.title = @"Synchronized Directive";
    cellModel4.actionBlock = ^{
        
        SynchronizedDirectiveSample *sample = [[SynchronizedDirectiveSample alloc] init];
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
//            [sample foo];
//            [sample bar];
        });
        
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
//            [sample foo];
//            [sample bar];
        });
        
        @synchronized (sample) {
            [sample testSynchronizedRecursive];
        }
        
        @try {
            [sample exceptionWithSynchronized];
        } @catch (NSException *exception) {
            NSLog(@"Exception %@", exception);
        }
    };
    
    CellModel *cellModel41 = [[CellModel alloc] init];
    cellModel41.title = @"Thread Run without start Runloop";
    cellModel41.actionBlock = ^{
        NSThread *thread = [[NSThread alloc] initWithBlock:^{
            NSLog(@"Thread execute!!!");
//            [[NSRunLoop currentRunLoop] run];
//            [[NSRunLoop currentRunLoop] runMode:NSRunLoopCommonModes beforeDate:[NSDate distantFuture]];
            NSLog(@"Runloop Exit!!!");
        }];
        [thread start];
        
        [self performSelector:@selector(testMethod)
                     onThread:thread
                   withObject:nil
                waitUntilDone:NO]; // YES

        [self performSelector:@selector(testMethod2)
                     onThread:thread
                   withObject:nil
                waitUntilDone:NO]; // YES
        
        
        NSLog(@"after perform selector");
    };
    
    __weak typeof(_tableView) wkTbView = _tableView;
    CellModel *cellModel42 = [[CellModel alloc] init];
    cellModel42.title = @"Runloop Observer";
    cellModel42.actionBlock = ^{
        [RunloopObserverSample startRunloopObserver];
    };
    
    CellModel *cellModel5 = [[CellModel alloc] init];
    cellModel5.title = @"Queue Vs Thread";
    cellModel5.actionBlock = ^{
        dispatch_queue_t serialQueue = dispatch_queue_create("TestSerialQueue", DISPATCH_QUEUE_SERIAL);
        
        dispatch_set_target_queue(serialQueue, dispatch_get_main_queue());
        dispatch_async(serialQueue, ^{
            NSLog(@"currenct Thread: %@", [NSThread currentThread]);
        });
        
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
            NSLog(@"currenct Thread: %@", [NSThread currentThread]);
        });
        
        dispatch_async(dispatch_get_main_queue(), ^{
            NSLog(@"currenct Thread: %@", [NSThread currentThread]);
        });
    };
    
    CellModel *cellModel6 = [[CellModel alloc] init];
    cellModel6.title = @"Queue Vs Thread";
    cellModel6.actionBlock = ^{
        dispatch_queue_t serialQueue = dispatch_queue_create("TestSerialQueue", DISPATCH_QUEUE_SERIAL);
        
        dispatch_set_target_queue(serialQueue, dispatch_get_main_queue());
        dispatch_async(serialQueue, ^{
            NSLog(@"currenct Thread: %@", [NSThread currentThread]);
        });
        
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
            NSLog(@"currenct Thread: %@", [NSThread currentThread]);
        });
    };
    
    [_cellModels addObjectsFromArray:@[cellModel1, cellModel2, cellModel3, cellModel4, cellModel41, cellModel42, cellModel5, cellModel6]];
}

- (void)viewDidLoad {
    [super viewDidLoad];
    // Do any additional setup after loading the view, typically from a nib.
    
    [self setupCellModels];
    
    _tableView = [[UITableView alloc] initWithFrame:self.view.bounds style:UITableViewStylePlain];
    _tableView.delegate = self;
    _tableView.dataSource = self;
    [self.view addSubview:_tableView];
    
    [_tableView registerClass:[UITableViewCell class] forCellReuseIdentifier:@"CellIdentifier"];
    
    _tableView.tableHeaderView = [[UIView alloc] initWithFrame:CGRectMake(0, 0, CGRectGetWidth(self.view.bounds), 60)];
    _tableView.tableFooterView = [UIView new];
    
    NSArray *hCon = [NSLayoutConstraint constraintsWithVisualFormat:@"H:|[_tableView]|" options:0 metrics:nil views:NSDictionaryOfVariableBindings(_tableView)];
    NSArray *vCon = [NSLayoutConstraint constraintsWithVisualFormat:@"V:|[_tableView]|" options:0 metrics:nil views:NSDictionaryOfVariableBindings(_tableView)];
    
    _tableView.translatesAutoresizingMaskIntoConstraints = NO;
    [self.view addConstraints:hCon];
    [self.view addConstraints:vCon];
}


- (void)didReceiveMemoryWarning {
    [super didReceiveMemoryWarning];
    // Dispose of any resources that can be recreated.
}

#pragma mark -
#pragma makr UITableViewDelegate

- (CGFloat)tableView:(UITableView *)tableView heightForRowAtIndexPath:(NSIndexPath *)indexPath
{
    return 60;
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
    [tableView deselectRowAtIndexPath:indexPath animated:NO];
    
    // do something you like
    _cellModels[indexPath.row].actionBlock ? _cellModels[indexPath.row].actionBlock() : nil;
}

#pragma mark -
#pragma mark UITableViewDatasource

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    return _cellModels.count;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:@"CellIdentifier" forIndexPath:indexPath];
    
    // 模拟卡顿
    usleep(400*1000);
    
    cell.textLabel.text = _cellModels[indexPath.row].title;
    
    return cell;
}

@end
