//
//  iOSConcurrencySample.m
//  iOSMultiThreadSample
//
//  Created by xiaoyong.cxy on 18/07/2017.
//  Copyright © 2017 maple. All rights reserved.
//

#import "iOSConcurrencySample.h"

#define HAS_JOB     1
#define NO_JOB      0

@interface iOSConcurrencySample () {
    NSCondition         *_condition;
    BOOL                _hasJob;
    
    NSConditionLock     *_conditionLock;
    
    NSLock              *_lockForCP;
    
    NSThread            *_producerThread;
    NSThread            *_cousumerThread;
    
}

+ (instancetype)sharedInstance;

@end

@implementation iOSConcurrencySample


+ (instancetype)sharedInstance
{
    static iOSConcurrencySample *gSample = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        gSample = [[iOSConcurrencySample alloc] init];
    });
    
    return gSample;
}

- (instancetype)init
{
    self = [super init];
    if (self) {
        _condition = [[NSCondition alloc] init];
        _hasJob = NO;
        
        _conditionLock = [[NSConditionLock alloc] initWithCondition:NO_JOB];
        
        _lockForCP = [[NSLock alloc] init];
    }
    
    return self;
}

#pragma mark -
#pragma mark Producer & Consumer

- (void)cancelExistThread
{
    [_producerThread cancel];
    [_cousumerThread cancel];
}

- (void)consumerWithCondition
{
    while (YES && ![NSThread currentThread].isCancelled) {
        [_condition lock];
        while (!_hasJob) {
            NSLog(@"has no job, wait for producer");
            [_condition wait];
        }
        
        NSLog(@"Consumer Job");
        _hasJob = NO;
        
//        [_condition broadcast];
        [_condition unlock];
    }
}

- (void)producerWithCondition
{
    while (YES && ![NSThread currentThread].isCancelled) {
        [_condition lock];
        while (_hasJob) {
            NSLog(@"job exist, wait for consumer");
            [_condition wait];
        }
        
        NSLog(@"Producer Job");
        _hasJob = YES;
        
        [_condition broadcast];
        [_condition unlock];
    }
}

- (void)_producerConsumerWithCondition
{
    _producerThread = [[NSThread alloc] initWithTarget:self selector:@selector(producerWithCondition) object:nil];
    _cousumerThread = [[NSThread alloc] initWithTarget:self selector:@selector(consumerWithCondition) object:nil];
    [_producerThread start];
    [_cousumerThread start];
}

+ (void)producerConsumerWithCondition
{
    [[iOSConcurrencySample sharedInstance] cancelExistThread];
    [[iOSConcurrencySample sharedInstance] _producerConsumerWithCondition];
}

- (void)producerWithConditionLock
{
    int jobIndex = 0;
    while (YES && ![NSThread currentThread].isCancelled) {
        [_conditionLock lockWhenCondition:NO_JOB];
        
        sleep(arc4random()%2);
        
        NSLog(@"Producer Job %d", jobIndex++);
        [_conditionLock unlockWithCondition:HAS_JOB];
    }
}

- (void)consumerWithConditionLock
{
    int jobIndex = 0;
    while (YES && ![NSThread currentThread].isCancelled) {
        [_conditionLock lockWhenCondition:HAS_JOB];
        
        NSLog(@"Consumer Job %d", jobIndex++);
        
        [_conditionLock unlockWithCondition:NO_JOB];
    }
}

- (void)_producerConsumerWithConditionLock
{
    _producerThread = [[NSThread alloc] initWithTarget:self selector:@selector(producerWithConditionLock) object:nil];
    _cousumerThread = [[NSThread alloc] initWithTarget:self selector:@selector(consumerWithConditionLock) object:nil];
    [_producerThread start];
    [_cousumerThread start];
}

+ (void)producerConsumerWithConditionLock
{
    [[iOSConcurrencySample sharedInstance] cancelExistThread];
    [[iOSConcurrencySample sharedInstance] _producerConsumerWithConditionLock];
}

- (void)producerWithLock
{
    int jobIndex = 0;
    while (YES && ![NSThread currentThread].isCancelled) {
        
        [_lockForCP lock];
        while (_hasJob) {
            [_lockForCP unlock];
            [_lockForCP lock];
        }
        
        _hasJob = YES;
        NSLog(@"Producer Job %d", jobIndex++);
        [_lockForCP unlock];
    }
}

- (void)consumerWithLock
{
    int jobIndex = 0;
    while (YES && ![NSThread currentThread].isCancelled) {
        [_lockForCP lock];
        while (!_hasJob) {
            [_lockForCP unlock];
            [_lockForCP lock];
        }
        
        NSLog(@"Consumer Job %d", jobIndex++);
        _hasJob = NO;
        
        [_lockForCP unlock];
    }
}

- (void)_producerConsumerWithLock
{
    _producerThread = [[NSThread alloc] initWithTarget:self selector:@selector(producerWithLock) object:nil];
    _cousumerThread = [[NSThread alloc] initWithTarget:self selector:@selector(consumerWithLock) object:nil];
    [_producerThread start];
    [_cousumerThread start];
}

+ (void)producerConsumerWithLock
{
    [[iOSConcurrencySample sharedInstance] cancelExistThread];
    [[iOSConcurrencySample sharedInstance] _producerConsumerWithLock];
}

@end
